/*
 * TeamSpeak Hello Audio World sample
 *
 * Copyright (c) 2007-2017 TeamSpeak Systems
 */

#include "common.h"

#include <teamspeak/public_definitions.h>
#include <teamspeak/clientlib_publicdefinitions.h>
#include <teamspeak/public_errors.h>
#include <teamspeak/clientlib.h>

#if defined(_MSC_VER) && !defined(_CRT_SECURE_NO_WARNINGS)
#define _CRT_SECURE_NO_WARNINGS
#endif

#ifndef _WIN32
#include <unistd.h>
#endif
#include <stdint.h>
#include <stdio.h>
#include <string.h>

/*A global variable that indicates the current connection status. Note that in
  a real world scenario, one would not use global variables.*/
int connection_status = STATUS_DISCONNECTED;

/*A global variable holding a pointer to the current TeamSpeak client identity*/
char* identity = NULL;

/*
* Callback for connection status change.
* Connection status switches through the states STATUS_DISCONNECTED, STATUS_CONNECTING, STATUS_CONNECTED and STATUS_CONNECTION_ESTABLISHED.
*
* Parameters:
*   connection_id             - Server connection handler ID
*   newStatus                 - New connection status, see the enum ConnectStatus in clientlib_publicdefinitions.h
*   errorNumber               - Error code. Should be zero when connecting or actively disconnection.
*                               Contains error state when losing connection.
*/
void onConnectStatusChangeEvent(uint64 connection_id, int new_status, unsigned int connection_error)
{
    printf("Connect status changed: %llu %d %u\n", (unsigned long long)connection_id, new_status, connection_error);

    if (new_status == STATUS_DISCONNECTED)
    {
        if (connection_error == ERROR_failed_connection_initialisation)
        {
            printf("Looks like there is no server running.\n");
        }
    }

    /*Update the global variable to the new state so the wait_for_connection
     function below can use it.*/
    connection_status = new_status;
}

/* Initialize client lib with callbacks */
void ts_init(const char* soundbackends_path)
{
    /* Create struct for callback function pointers */
    struct ClientUIFunctions funcs;

    /* Initialize all callbacks with NULL */
    memset(&funcs, 0, sizeof(struct ClientUIFunctions));

    /* Callback function pointers */
    /* It is sufficient to only assign those callback functions you are using. When adding more callbacks, add those function pointers here. */
    funcs.onConnectStatusChangeEvent = onConnectStatusChangeEvent;

    CHECK_ERROR_FATAL(ts3client_initClientLib(&funcs, NULL, LogType_CONSOLE, NULL, soundbackends_path));

    /* Create a new client identity */
    /* In your real application you should do this only once, store the assigned identity locally and then reuse it. */
    CHECK_ERROR_FATAL(ts3client_createIdentity(&identity));
}

/*Create new server connection handler and connect it to the server, return connection id*/
uint64 ts_connect()
{
    uint64 connection_id = 0;
    /* Spawn a new server connection handler using the default port and store the server ID */
    CHECK_ERROR_FATAL(ts3client_spawnNewServerConnectionHandler(0, &connection_id));

    /* Connect to server on localhost:9987 with nickname "client", no default channel, no default channel password and server password "secret" */
    CHECK_ERROR_FATAL(ts3client_startConnection(connection_id, identity, "localhost", 9987, "client", NULL, "", "secret"));

    return connection_id;
}

/*A function to put the current thread into a wait loop until the ts3 connection
  has either fully connected, or there was a connection error (signaled by a 
  disconnect.
  returns 0 on error, 1 on connection fully established.
  */
bool wait_for_connection()
{
    /*Wait for connection_status to be "STATUS_CONNECTION_ESTABLISHED" which means fully connected */
    while (connection_status != STATUS_CONNECTION_ESTABLISHED)
    {
        /*If there was a connection error, the state will go to "STATUS_DISCONNECTED" so it is needed to check for that*/
        if (connection_status == STATUS_DISCONNECTED)
        {
            /*connection error occurred.*/
            return false;
        }
        sleep_ms(20);
    }

    return true;
}

/*Free used memory and shutdown the connection*/
void ts_shutdown()
{
    ts3client_freeMemory(identity);  /* Release dynamically allocated memory */
    identity = NULL;

    /* Shutdown client lib */
    CHECK_ERROR_FATAL(ts3client_destroyClientLib());
}

void start_audio(uint64 connection_id)
{
    /* Open default playback device */
    /* Passing empty string for mode and NULL or empty string for device will open the default device */
    CHECK_ERROR_FATAL(ts3client_openPlaybackDevice(connection_id, "", NULL));

    /* Open default capture device */
    /* Passing empty string for mode and NULL or empty string for device will open the default device */
    CHECK_ERROR_FATAL(ts3client_openCaptureDevice(connection_id, "", NULL));
}

void stop_audio(uint64 connection_id)
{
    // Close the default capture device
    CHECK_ERROR_FATAL(ts3client_closeCaptureDevice(connection_id));

    // Close the default playback device
    CHECK_ERROR_FATAL(ts3client_closePlaybackDevice(connection_id));
}

bool is_dsp_enabled(uint64 connection_id, const char* dsp)
{
    char* enabled;
    if (CHECK_ERROR(ts3client_getPreProcessorConfigValue(connection_id, dsp, &enabled)))
    {
        const bool val = strcmp("true", enabled) == 0;
        ts3client_freeMemory(enabled);
        return val;
    }
    return false;
}

void toggle_dsp(uint64 connection_id, const char* dsp)
{
    const bool target_enable = !is_dsp_enabled(connection_id, dsp);
    if (CHECK_ERROR(ts3client_setPreProcessorConfigValue(connection_id, dsp, target_enable ? "true" : "false")))
        printf("\nToggled %s %s.\n\n", dsp, target_enable ? "on" : "off");
}

void set_vad_level(uint64 connection_id)
{
    int vad, n;
    char s[100];

    printf("\nEnter VAD level: ");
    n = scanf("%d", &vad);
    emptyInputBuffer();
    if (n == 0)
    {
        printf("Invalid input. Please enter a number.\n\n");
        return;
    }

    /* Adjust "voiceactivation_level" preprocessor value */
    snprintf(s, 100, "%d", vad);
    CHECK_ERROR(ts3client_setPreProcessorConfigValue(connection_id, "voiceactivation_level", s));
    printf("\nSet VAD level to %s.\n\n", s);
}

void toggle_input_status(uint64 connection_id)
{
    int status;
    if (CHECK_ERROR(ts3client_getClientSelfVariableAsInt(connection_id, CLIENT_INPUT_DEACTIVATED, &status)))
    {
        status = (status == INPUT_ACTIVE) ? INPUT_DEACTIVATED : INPUT_ACTIVE;
        CHECK_ERROR(ts3client_setClientSelfVariableAsInt(connection_id, CLIENT_INPUT_DEACTIVATED, status));
    }
}

void stream_continuously(uint64 connection_id)
{
    CHECK_ERROR(ts3client_setPreProcessorConfigValue(connection_id, "vad", "false"));
    CHECK_ERROR(ts3client_setClientSelfVariableAsInt(connection_id, CLIENT_INPUT_DEACTIVATED, INPUT_ACTIVE));
    CHECK_ERROR(ts3client_flushClientSelfUpdates(connection_id, NULL));
}

void toggle_input_muted(uint64 connection_id)
{
    int status;
    if (CHECK_ERROR(ts3client_getClientSelfVariableAsInt(connection_id, CLIENT_INPUT_MUTED, &status)))
    {
        status = (status == MUTEINPUT_MUTED) ? MUTEINPUT_NONE : MUTEINPUT_MUTED;
        CHECK_ERROR(ts3client_setClientSelfVariableAsInt(connection_id, CLIENT_INPUT_MUTED, status));
        CHECK_ERROR(ts3client_flushClientSelfUpdates(connection_id, NULL));
    }
}

void toggle_mute_all(uint64 connection_id)
{
    anyID* visible_clients;
    if (CHECK_ERROR(ts3client_getClientList(connection_id, &visible_clients)))
    {
        // base the target status on the status of the first client in the array
        if (visible_clients[0] != 0)
        {
            int status = 0;
            CHECK_ERROR(ts3client_getClientVariableAsInt(connection_id, visible_clients[0], CLIENT_IS_MUTED, &status));
            if (status == 0)
            {
                CHECK_ERROR(ts3client_requestMuteClients(connection_id, visible_clients, NULL));
            }
            else
            {
                CHECK_ERROR(ts3client_requestUnmuteClients(connection_id, visible_clients, NULL));
            }
        }
        ts3client_freeMemory(visible_clients);
    }
}

void print_help()
{
    printf("\n--- Enter [q] to disconnect from server and exit ---\n");
    printf("[v] - Toggle Voice Activity Detection\n[V] - Set Voice Activity Detection level\n[t] - Toggle input status\n");
    printf("[a] - Toggle Automatic Gain Correction\n[d] - Toggle Denoiser\n[e] - Toggle Echo Cancellation unit\n");
}

int main(int argc, char **argv)
{
    UNUSED(argc);

    /* Resource path points to the SDK\bin directory to locate the soundbackends*/
    ts_init(program_path(argv[0]));

    const uint64 connection_id = ts_connect();
    if (connection_id != 0)
    {
        start_audio(connection_id);

        if (wait_for_connection())
        {
            /* if we get here, we are fully connected */
            print_help();
            /* Wait for user input */
            while (connection_status != STATUS_DISCONNECTED)
            {
                int c = getc(stdin);
                switch (c)
                {
                    case 'q':
                    {
                        printf("\nDisconnecting from server...\n");
                        CHECK_ERROR(ts3client_stopConnection(connection_id, "leaving"));
                        break;
                    }
                    case 'v':
                        toggle_dsp(connection_id, "vad");
                        break;
                    case 'V':
                        set_vad_level(connection_id);
                        break;
                    case 'a':
                        toggle_dsp(connection_id, "agc");
                        break;
                    case 'e':
                        toggle_dsp(connection_id, "echo_canceling");
                        break;
                    case 'd':
                        toggle_dsp(connection_id, "denoise");
                        break;
                    case 't':
                        toggle_input_status(connection_id);
                        break;
                    case 'm':
                        toggle_input_muted(connection_id);
                        break;
                    case 'M':
                        toggle_mute_all(connection_id);
                    default:
                        break;
                }
        
                sleep_ms(20);
            }
        }

        stop_audio(connection_id);
        /* Destroy server connection handler */
        CHECK_ERROR_FATAL(ts3client_destroyServerConnectionHandler(connection_id));
    }

    ts_shutdown();
    return 0;
}
